@@ -9,14 +9,14 @@ App::Ack - A container for functions for the ack program
 
 =head1 VERSION
 
-Version 1.69_01
+Version 1.75_01
 
 =cut
 
 our $VERSION;
 our $COPYRIGHT;
 BEGIN {
-    $VERSION = '1.69_01';
+    $VERSION = '1.75_01';
     $COPYRIGHT = 'Copyright 2005-2007 Andy Lester, all rights reserved.';
 }
 
@@ -36,23 +36,31 @@ use Getopt::Long ();
 
 BEGIN {
     %ignore_dirs = (
-        '.git'  => 'Git',
-        '.pc'   => 'quilt',
-        '.svn'  => 'Subversion',
-        CVS     => 'CVS',
-        RCS     => 'RCS',
-        SCCS    => 'SCCS',
-        _darcs  => 'darcs',
-        blib    => 'Perl module building',
-        'autom4te.cache'    => 'autoconf cache',
-        'cover_db'          => 'Devel::Cover results',
+        '.bzr'              => 'Bazaar',
+        '.cdv'              => 'Codeville',
+        '~.dep'             => 'Interface Builder',
+        '~.dot'             => 'Interface Builder',
+        '~.nib'             => 'Interface Builder',
+        '~.plst'            => 'Interface Builder',
+        '.git'              => 'Git',
+        '.hg'               => 'Mercurial',
+        '.pc'               => 'quilt',
+        '.svn'              => 'Subversion',
+        blib                => 'Perl module building',
+        CVS                 => 'CVS',
+        RCS                 => 'RCS',
+        SCCS                => 'SCCS',
+        _darcs              => 'darcs',
+        _sgbak              => 'Vault/Fortress',
+        'autom4te.cache'    => 'autoconf',
+        'cover_db'          => 'Devel::Cover',
     );
 
     %mappings = (
         asm         => [qw( s )],
         binary      => q{Binary files, as defined by Perl's -B op (default: off)},
         cc          => [qw( c h xs )],
-        cpp         => [qw( cpp m h )],
+        cpp         => [qw( cpp cc m hpp hh h )],
         csharp      => [qw( cs )],
         css         => [qw( css )],
         elisp       => [qw( el )],
@@ -87,7 +95,6 @@ BEGIN {
         xml         => [qw( xml dtd xslt )],
     );
 
-
     while ( my ($type,$exts) = each %mappings ) {
         if ( ref $exts ) {
             for my $ext ( @{$exts} ) {
@@ -152,8 +159,7 @@ sub get_command_line_options {
 
     my $getopt_specs = {
         1                       => sub { $opt{1} = $opt{m} = 1 },
-        a                       => \$opt{all},
-        'all!'                  => \$opt{all},
+        'a|all-types'           => \$opt{all},
         'A|after-context=i'     => \$opt{after_context},
         'B|before-context=i'    => \$opt{before_context},
         'C|context:i'           => sub { shift; my $val = shift; $opt{before_context} = $opt{after_context} = ($val || 2) },
@@ -167,6 +173,7 @@ sub get_command_line_options {
         'h|no-filename'         => \$opt{h},
         'H|with-filename'       => \$opt{H},
         'i|ignore-case'         => \$opt{i},
+        'lines=s'               => sub { shift; my $val = shift; push @{$opt{lines}}, $val },
         'l|files-with-matches'  => \$opt{l},
         'L|files-without-match' => sub { $opt{l} = $opt{v} = 1 },
         'm|max-count=i'         => \$opt{m},
@@ -176,6 +183,7 @@ sub get_command_line_options {
         'passthru'              => \$opt{passthru},
         'Q|literal'             => \$opt{Q},
         'sort-files'            => \$opt{sort_files},
+        'u|unrestricted'        => \$opt{u},
         'v|invert-match'        => \$opt{v},
         'w|word-regexp'         => \$opt{w},
 
@@ -212,11 +220,66 @@ sub get_command_line_options {
     Getopt::Long::GetOptions( %{$getopt_specs} ) or
         App::Ack::die( 'See ack --help or ack --man for options.' );
 
-    apply_defaults(\%opt);
+    my %defaults = (
+        all            => 0,
+        color          => $to_screen && !$App::Ack::is_windows,
+        follow         => 0,
+        group          => $to_screen,
+        before_context => 0,
+        after_context  => 0,
+    );
+    while ( my ($key,$value) = each %defaults ) {
+        if ( not defined $opt{$key} ) {
+            $opt{$key} = $value;
+        }
+    }
+
+    if ( defined $opt{m} && $opt{m} <= 0 ) {
+        App::Ack::die( '-m must be greater than zero' );
+    }
+
+    for ( qw( before_context after_context ) ) {
+        if ( defined $opt{$_} && $opt{$_} < 0 ) {
+            App::Ack::die( "--$_ may not be negative" );
+        }
+    }
 
     if ( defined( my $val = $opt{output} ) ) {
         $opt{output} = eval qq[ sub { "$val" } ];
     }
+    if ( defined( my $l = $opt{lines} ) ) {
+        # --line=1 --line=5 is equivalent to --line=1,5
+        my @lines = split( /,/, join( ',', @{$l} ) );
+
+        # --line=1-3 is equivalent to --line=1,2,3
+        @lines = map {
+            my @ret;
+            if ( /-/ ) {
+                my ($from, $to) = split /-/, $_;
+                if ( $from > $to ) {
+                    App::Ack::warn( "ignoring --line=$from-$to" );
+                    @ret = ();
+                }
+                else {
+                    @ret = ( $from .. $to );
+                }
+            }
+            else {
+                @ret = ( $_ );
+            };
+            @ret
+        } @lines;
+
+        if ( @lines ) {
+            my %uniq;
+            @uniq{ @lines } = ();
+            $opt{lines} = [ sort { $a <=> $b } keys %uniq ];   # numerical sort and each line occurs only once!
+        }
+        else {
+            # happens if there are only ignored --line directives
+            App::Ack::die( 'All --line options are invalid.' );
+        }
+    }
 
     return %opt;
 }
@@ -294,7 +357,7 @@ sub filetypes {
         return ('shell',TEXT)  if $header =~ /\b(?:ba|c|k|z)?sh\b/;
     }
     else {
-        return ('xml',TEXT)    if $header =~ /\Q<?xml /;
+        return ('xml',TEXT)    if $header =~ /\Q<?xml /i;
     }
 
     return (TEXT);
@@ -305,13 +368,20 @@ sub filetypes {
 Returns true if the filename is one that we can search, and false
 if it's one that we should ignore like a coredump or a backup file.
 
+Recognized files:
+  /~$/            - Unix backup files
+  /#.+#$/         - Emacs swap files
+  /[._].*\.swp$/  - Vi(m) swap files
+  /core\.\d+$/    - core dumps
+
 =cut
 
 sub is_searchable {
     my $filename = shift;
 
+    # If these are updated, update the --help message
     return if $filename =~ /~$/;
-    return if $filename =~ m{$path_sep_regex?(?:#.+#|core\.\d+)$}o;
+    return if $filename =~ m{$path_sep_regex?(?:#.+#|core\.\d+|[._].*\.swp)$}o;
 
     return 1;
 }
@@ -423,6 +493,7 @@ Searching:
   -Q, --literal         Quote all metacharacters; expr is literal
 
 Search output:
+  --line=NUM            Only print line(s) NUM of each file
   -l, --files-with-matches
                         Only print filenames containing matches
   -L, --files-without-match
@@ -432,7 +503,8 @@ Search output:
   --passthru            Print all lines, whether matching or not
   --output=expr         Output the evaluation of expr for each line
                         (turns off text highlighting)
-  -m, --max-count=NUM   Stop searching in a file after NUM matches
+  -m, --max-count=NUM   Stop searching in each file after NUM matches
+  -1                    Stop searching after one match of any kind
   -H, --with-filename   Print the filename for each match
   -h, --no-filename     Suppress the prefixing filename on output
   -c, --count           Show number of lines matching per file
@@ -461,9 +533,9 @@ File finding:
   --sort-files          Sort the found files lexically.
 
 File inclusion/exclusion:
+  -a, --all-types       All file types searched; directories still skipped
+  -u, --unrestricted    All files and directories searched
   -n                    No descending into subdirectories
-  -a, --all             All files, regardless of extension (but still skips
-                        $ignore_dirs dirs)
   --perl                Include only Perl files.
   --type=perl           Include only Perl files.
   --noperl              Exclude Perl files.
@@ -471,6 +543,15 @@ File inclusion/exclusion:
                         See "ack --help type" for supported filetypes.
   --[no]follow          Follow symlinks.  Default is off.
 
+  Directories ignored by default:
+    $ignore_dirs
+
+  Files not checked for type:
+    /~\$/           - Unix backup files
+    /#.+#\$/        - Emacs swap files
+    /[._].*\\.swp\$/ - Vi(m) swap files
+    /core\\.\\d+\$/   - core dumps
+
 Miscellaneous:
   --help                This help
   --man                 Man page
@@ -521,7 +602,12 @@ sub _listify {
     return '' if !@whats;
 
     my $end = pop @whats;
-    return @whats ? join( ', ', @whats ) . " and $end" : $end;
+    my $str = @whats ? join( ', ', @whats ) . " and $end" : $end;
+
+    no warnings 'once';
+    require Text::Wrap;
+    $Text::Wrap::columns = 75;
+    return Text::Wrap::wrap( '', '    ', $str );
 }
 
 =head2 get_version_statement( $copyright )
@@ -651,6 +737,36 @@ sub close_file {
     return 0;
 }
 
+
+=head2 needs_line_scan( $fh, $regex )
+
+Slurp up an entire file up to 100K, see if there are any matches
+in it, and if so, let us know so we can iterate over it directly.
+If it's bigger than 100K, we have to do the line-by-line, too.
+
+=cut
+
+sub needs_line_scan {
+    my $fh = shift;
+    my $regex = shift;
+
+    my $size = -s $fh;
+
+    if ( $size > 100_000 ) {
+        return 1;
+    }
+
+    my $buffer;
+    my $rc = sysread( $fh, $buffer, $size );
+    if ( defined($rc) && ( $rc == $size ) && ( $buffer =~ /$regex/osm ) ) {
+        return 1;
+    }
+    else {
+        return 0;
+    }
+}
+
+
 =head2 search( $fh, $could_be_binary, $filename, $regex, \%opt )
 
 Main search method
@@ -662,10 +778,6 @@ Main search method
     my $regex;
     my $display_filename;
 
-    my $output_func;
-    my $show_filename;
-    my $group;
-    my $color;
     my $keep_context;
 
     my $last_output_line; # number of the last line that has been output
@@ -685,10 +797,15 @@ sub search {
     my $nmatches = 0;
 
     $display_filename = undef;
-    $output_func = $opt->{output};
-    $show_filename = $opt->{show_filename};
-    $group = $opt->{group};
-    $color = $opt->{color};
+
+    # for --line processing
+    my $has_lines = 0;
+    my @lines;
+    if ( defined $opt->{lines} ) {
+        $has_lines = 1;
+        @lines = ( @{$opt->{lines}}, -1 );
+        undef $regex; # Don't match when printing matching line
+    }
 
     # for context processing
     $last_output_line = -1;
@@ -699,24 +816,43 @@ sub search {
     $keep_context = ($before_context || $after_context) && !$passthru;
 
     my @before;
+    my $before_starts_at_line;
     my $after = 0; # number of lines still to print after a match
 
     while (<$fh>) {
-        if ( $v ? /$regex/o : !/$regex/o ) {
-            print if $passthru;
+        # XXX Optimize away the case when there are no more @lines to find.
+        if ( $has_lines
+               ? $. != $lines[0]  # $lines[0] should be a scalar
+               : $v ? /$regex/o : !/$regex/o ) {
+            if ( $passthru ) {
+                print;
+                next;
+            }
+
             if ( $keep_context ) {
                 if ( $after ) {
-                    print_match_or_context( $_, $., 0 );
+                    print_match_or_context( $opt, 0, $., $_ );
                     $after--;
                 }
-                else {
-                    push @before, [$_, $.];
-                    shift @before if @before > $before_context;
+                elsif ( $before_context ) {
+                    if ( @before ) {
+                        if ( @before >= $before_context ) {
+                            shift @before;
+                            ++$before_starts_at_line;
+                        }
+                    }
+                    else {
+                        $before_starts_at_line = $.;
+                    }
+                    push @before, $_;
                 }
+                last if $max && ( $nmatches >= $max ) && !$after;
             }
             next;
-        } 
+        } # not a match
+
         ++$nmatches;
+        shift @lines if $has_lines;
 
         if ( $could_be_binary ) {
             if ( -B $filename ) {
@@ -726,18 +862,24 @@ sub search {
             $could_be_binary = 0;
         }
         if ( $keep_context ) {
-            for my $i ( @before ) {
-                print_match_or_context( @{$i}[0,1], 0 );
+            if ( @before ) {
+                print_match_or_context( $opt, 0, $before_starts_at_line, @before );
+                @before = ();
+                $before_starts_at_line = 0;
+            }
+            if ( $max && $nmatches > $max ) {
+                --$after;
+            }
+            else {
+                $after = $after_context;
             }
-            @before = ();
-            $after = $after_context;
         }
-        print_match_or_context( $_, $., 1 );
+        print_match_or_context( $opt, 1, $., $_ );
 
-        last if $max && ( $nmatches >= $max );
+        last if $max && ( $nmatches >= $max ) && !$after;
     } # while
 
-    if ( $nmatches && $show_filename && $group ) {
+    if ( $nmatches && $opt->{show_filename} && $opt->{group} ) {
         print "\n";
     }
 
@@ -745,17 +887,20 @@ sub search {
 }   # search()
 
 
-=head2 print_match_or_context( $line, $line_no, $is_match )
+=head2 print_match_or_context( $opt, $is_match, $starting_line_no, @lines )
 
 Prints out a matching line or a line of context around a match.
 
 =cut
 
 sub print_match_or_context {
-    local $_     = shift; # line to print
-    my $line_no  = shift; # line number of that line
+    my $opt      = shift; # opts array
     my $is_match = shift; # is there a match on the line?
-    my $sep = $is_match ? ':' : '-';
+    my $line_no  = shift;
+
+    my $color = $opt->{color};
+    my $group = $opt->{group};
+    my $show_filename = $opt->{show_filename};
 
     if ( $show_filename ) {
         if ( not defined $display_filename ) {
@@ -763,41 +908,48 @@ sub print_match_or_context {
                 $color
                     ? Term::ANSIColor::colored( $filename, $ENV{ACK_COLOR_FILENAME} )
                     : $filename;
-        }
-        if ( $group && !$any_output ) {
-            print $display_filename, "\n";
-        }
-    }
-
-    if ( $keep_context && !$output_func ) {
-        if ( ( $last_output_line != $line_no - 1 ) &&
-            ( $any_output || ( !$group && $context_overall_output_count++ > 0 ) ) ) {
-            print "--\n";
-        }
-        # to ensure separators between different files when --nogroup
-
-        $last_output_line = $line_no;
-    }
-    $any_output = 1;
-
-    if ( $show_filename ) {
-        print $display_filename, $sep unless $group;
-        print $line_no, $sep;
-    }
-
-    if ( $output_func ) {
-        while ( /$regex/go ) {
-            print $output_func->(), "\n";
-        }
-    }
-    else {
-        if ( $color && $is_match ) {
-            if ( s/($regex)/Term::ANSIColor::colored($1,$ENV{ACK_COLOR_MATCH})/eg ) {
-                s/\n$/\e[0m\e[K\n/;     # Before \n, reset the color and clear to end of line
+            if ( $group && !$any_output ) {
+                print $display_filename, "\n";
             }
         }
-        print;
     }
+
+    my $sep = $is_match ? ':' : '-';
+    my $output_func = $opt->{output};
+    for ( @_ ) {
+        if ( $keep_context && !$output_func ) {
+            if ( ( $last_output_line != $line_no - 1 ) &&
+                ( $any_output || ( !$group && $context_overall_output_count++ > 0 ) ) ) {
+                print "--\n";
+            }
+            # to ensure separators between different files when --nogroup
+
+            $last_output_line = $line_no;
+        }
+
+        if ( $show_filename ) {
+            print $display_filename, $sep if not $group;
+            print $line_no, $sep;
+        }
+
+        if ( $output_func ) {
+            while ( /$regex/go ) {
+                print $output_func->(), "\n";
+            }
+        }
+        else {
+            if ( $color && $is_match && $regex ) {
+                if ( s/($regex)/Term::ANSIColor::colored($1,$ENV{ACK_COLOR_MATCH})/eg ) {
+                    s/\n$/\e[0m\e[K\n/;     # Before \n, reset the color and clear to end of line
+                }
+            }
+            print;
+        }
+        $any_output = 1;
+        ++$line_no;
+    }
+
+    return;
 } # print_match_or_context()
 
 } # scope around search() and print_match_or_context()
@@ -851,33 +1003,6 @@ sub search_and_list {
 }   # search_and_list()
 
 
-=head2 apply_defaults
-
-Apply the default options
-
-=cut
-
-sub apply_defaults {
-    my $opt = shift;
-
-    my %defaults = (
-        all            => 0,
-        color          => $to_screen && !$App::Ack::is_windows,
-        follow         => 0,
-        group          => $to_screen,
-        m              => 0,
-        before_context => 0,
-        after_context  => 0,
-    );
-    while ( my ($key,$value) = each %defaults ) {
-        if ( not defined $opt->{$key} ) {
-            $opt->{$key} = $value;
-        }
-    }
-
-    return;
-}
-
 =head2 filetypes_supported_set
 
 True/False - are the filetypes set?
